﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text.RegularExpressions;
using System.Xml.Linq;
using NTERA.Console.RenderItem;
using NTERA.Core;
using NTERA.Core.Interop;
using NTERA.EmuEra.Game.EraEmu.Content;

namespace NTERA.Console
{
	public static class HtmlParser
	{
		public static IEnumerable<IRenderItem> ParseHtml(string html, Func<string, CroppedImage> assetFunc)
		{
			//fix broken HTML from shitty emuera format
			string fixedHtml = Regex.Replace(html, @"<img([^\/]*?)>", "<img$1 />");
			fixedHtml = fixedHtml.Replace("<br>", "<br/>");
			fixedHtml = fixedHtml.Replace("<nobr>", "<nobr/>");

			var element = XElement.Parse($"<parent>{fixedHtml}</parent>");

			return ParseHtml_Internal(element, assetFunc, HtmlStyle.Default);
		}

		private static IEnumerable<IRenderItem> ParseHtml_Internal(XElement xmlNode, Func<string, CroppedImage> assetFunc, HtmlStyle style)
		{
			List<IRenderItem> renderItems = new List<IRenderItem>();
			HtmlStyle localStyle = (HtmlStyle)style.Clone();

			foreach (var node in xmlNode.Elements())
			{
				switch (node.Name.LocalName)
				{
					case "p":
						string alignment = node.Attribute("align")?.Value;

						if (alignment != null)
							localStyle.Alignment = (DisplayLineAlignment)Enum.Parse(typeof(DisplayLineAlignment), alignment.ToUpper());

						renderItems.AddRange(ParseHtml_Internal(node, assetFunc, localStyle));
						break;

					case "img":
						string src = node.Attribute("src")?.Value;

						if (src == null)
							throw new InvalidOperationException("HTML 'img' tag in script is missing 'src' attribute");

						var image = assetFunc(src);

						renderItems.Add(new ImageRenderItem(image.Bitmap, image.Rectangle, alignment: localStyle.Alignment));
						break;

					case "br":
						renderItems.Add(new TextRenderItem(""));
						break;

					case "nobr":
						//renderItems.Add(new TextRenderItem(""));
						break;

					case "font":
						string color = node.Attribute("color")?.Value;

						if (color != null)
						{
							if (color.StartsWith("#"))
							{
								string colorHex = color.Substring(1).PadLeft(8, 'F');

								localStyle.ForeColor = Color.FromArgb(Convert.ToInt32(colorHex, 16));
							}
							else
								localStyle.ForeColor = Color.FromName(color);
						}

						renderItems.AddRange(ParseHtml_Internal(node, assetFunc, localStyle));
						break;

					case "nonbutton":
						//renderItems.Add(new TextRenderItem(node.Value, color: localStyle.ForeColor, alignment: localStyle.Alignment));
						renderItems.AddRange(ParseHtml_Internal(node, assetFunc, localStyle));
						break;

					default:
						renderItems.Add(new TextRenderItem(node.ToString(), color: localStyle.ForeColor, alignment: localStyle.Alignment));
						break;
				}
			}

			if (!xmlNode.HasElements && !string.IsNullOrWhiteSpace(xmlNode.Value))
				renderItems.Add(new TextRenderItem(xmlNode.Value, color: localStyle.ForeColor, alignment: localStyle.Alignment));

			return renderItems;
		}

		private class HtmlStyle : ICloneable
		{
			public static HtmlStyle Default => new HtmlStyle();

			public DisplayLineAlignment Alignment = DisplayLineAlignment.LEFT;

			public Color ForeColor = Color.White;

			public object Clone()
			{
				return MemberwiseClone();
			}
		}
	}
}
